import { container_paths, containers_path, state_config_path } from "./paths.ts";
import { log_from } from "./utils.ts";
import { z } from "https://deno.land/x/zod@v3.22.4/mod.ts";

const log = log_from("config");

export type StateConfig = ReturnType<typeof new_config>;
export function new_config(app_key: string, app_secret: string, consumer_key: string) {
	return { app_key, app_secret, consumer_key };
}

export async function load_state_config() {
	try {
		const content = await Deno.readTextFile(state_config_path());
		return JSON.parse(content) as StateConfig;
	} catch (_) {
		const result = new_config("APP_KEY", "APP_SECRET", "CONSUMER_KEY");
		const content = JSON.stringify(result);
		await Deno.writeTextFile(state_config_path(), content);
		return result;
	}
}

const CONTAINER_CONFIG_VERSION = 1;

export class ContainerConfig {
	name;
	redirections;

	constructor(name: string, redirections: ContainerConfigRedirection[]) {
		this.name = name;
		this.redirections = redirections;
	}

	public static new_default(name: string) {
		return new ContainerConfig(name, []);
	}

	public static async load(name: string) {
		const path = container_paths(name).configuration;
		const content = await Deno.readTextFile(path);
		return ContainerConfig.deserialize(content);
	}

	public async save() {
		const path = container_paths(this.name).configuration;
		await Deno.writeTextFile(path, this.serialize());
	}

	public static async *load_all() {
		for await (const { isDirectory, name } of Deno.readDir(containers_path())) {
			if (!isDirectory) continue;
			yield await ContainerConfig.load(name);
		}
	}

	public static deserialize(raw: string) {
		const unknown = JSON.parse(raw);
		const parsed = parse(unknown);
		return new ContainerConfig(parsed.name, parsed.redirections);
	}

	public serialize() {
		const raw: SerializedContainerConfig = {
			version: CONTAINER_CONFIG_VERSION,
			name: this.name,
			redirections: this.redirections,
		};
		return JSON.stringify(raw);
	}

	public *used_host_ports() {
		for (const redir of this.redirections) {
			if (redir.kind === "http") yield redir.port;
			if (redir.kind === "port") yield redir.from;
		}
	}
}

function parse(input: unknown) {
	const redir_port = z.object({
		kind: z.literal("port"),
		from: z.number(),
		to: z.number(),
	});

	const redir_http = z.object({
		kind: z.literal("http"),
		tls: z.boolean(),
		port: z.number(),
		domain: z.string(),
	});

	// TODO : redir DNS SRV

	const redirection = redir_http.or(redir_port);

	return z.object({
		name: z.string(),
		version: z.number(),
		redirections: z.array(redirection),
	}).parse(input);
}

type SerializedContainerConfig = ReturnType<typeof parse>;
export type ContainerConfigRedirection = SerializedContainerConfig["redirections"][0];
